
#include <windows.h>
#include <setupapi.h>
#include <stdio.h>



HANDLE
OpenViaInterface(
                 LPGUID	pInterfaceGuid,
                 ULONG InstanceIndex,
                 DWORD dwDesiredAccess,                      // access mode
                 DWORD dwShareMode,                          // share mode
                 LPSECURITY_ATTRIBUTES lpSecurityAttributes, // SD
                 DWORD dwCreationDisposition,                // how to create
                 DWORD dwFlagsAndAttributes,                 // file attributes
                 HANDLE hTemplateFile                        // handle to template file
                 )
                 /*++
                 Routine Description:



                 Arguments:


                 Return Value:

                 return HANDLE if the open and initialization was successful,
                 else INVLAID_HANDLE_VALUE.

                 If there is no such device then error code ERROR_NO_MORE_ITEMS
                 is set. Call GetLastError to retrieve the error code.
                 --*/
{
    HANDLE						hOut = INVALID_HANDLE_VALUE;
    HDEVINFO					hardwareDeviceInfo;
    SP_INTERFACE_DEVICE_DATA	deviceInfoData;
    PSP_INTERFACE_DEVICE_DETAIL_DATA	functionClassDeviceData = NULL;
    ULONG						predictedLength;
    ULONG						requiredLength;
    DWORD						err=ERROR_SUCCESS;


    //
    // Open a handle to the plug and play dev node.
    // SetupDiGetClassDevs() returns a device information set that contains info on all
    // installed devices of a specified class.
    //
    hardwareDeviceInfo = SetupDiGetClassDevs (
        pInterfaceGuid,
        NULL, // Define no enumerator (global)
        NULL, // Define no
        (DIGCF_PRESENT | // Only Devices present
        DIGCF_INTERFACEDEVICE)); // Function class devices.
    if(hardwareDeviceInfo==INVALID_HANDLE_VALUE)
    {
        err=GetLastError();
        goto ret;
    }



    // SetupDiEnumDeviceInterfaces() returns information about device interfaces
    // exposed by one or more devices. Each call returns information about one interface;
    // the routine can be called repeatedly to get information about several interfaces
    // exposed by one or more devices.

    deviceInfoData.cbSize = sizeof (SP_INTERFACE_DEVICE_DATA);
    if (!SetupDiEnumDeviceInterfaces (hardwareDeviceInfo,
        0, // We don't care about specific PDOs
        pInterfaceGuid,
        InstanceIndex,
        &deviceInfoData))
    {
        err=GetLastError();
        goto ret;
    }


    //
    // Allocate a function class device data structure to receive the
    // goods about this particular device.
    //
    requiredLength = 0;
    SetupDiGetInterfaceDeviceDetail (
        hardwareDeviceInfo,
        &deviceInfoData,
        NULL, // probing so no output buffer yet
        0, // probing so output buffer length of zero
        &requiredLength,
        NULL); // not interested in the specific dev-node


    predictedLength=requiredLength;

    functionClassDeviceData = (PSP_INTERFACE_DEVICE_DETAIL_DATA)malloc(predictedLength);
    if (!functionClassDeviceData)
    {
        err=GetLastError();
        goto ret;
    }

    //
    // Retrieve the information from Plug and Play.
    //
    functionClassDeviceData->cbSize = sizeof (SP_INTERFACE_DEVICE_DETAIL_DATA);
    if (! SetupDiGetInterfaceDeviceDetail (
        hardwareDeviceInfo,
        &deviceInfoData,
        functionClassDeviceData,
        predictedLength,
        &requiredLength,
        NULL))
    {
        err=GetLastError();
        goto ret;
    }

    //
    // Open connection
    //

    hOut = CreateFile (
        functionClassDeviceData->DevicePath,
        dwDesiredAccess,                      // access mode
        dwShareMode,                          // share mode
        lpSecurityAttributes, // SD
        dwCreationDisposition,                // how to create
        dwFlagsAndAttributes,                 // file attributes
        hTemplateFile                        // handle to template file
        );

    if (hOut==INVALID_HANDLE_VALUE)
    {
        err=GetLastError();
        fprintf(stderr, "FAILED to open %s, error=%lu\n", functionClassDeviceData->DevicePath, (long unsigned)err );
        goto ret;
    }


ret:
    //
    // Free allocated resources
    //
    if (functionClassDeviceData)
    {
        free(functionClassDeviceData);
    }
    if (hardwareDeviceInfo!=INVALID_HANDLE_VALUE)
    {
        SetupDiDestroyDeviceInfoList(hardwareDeviceInfo);
    }
    //
    // Set error code
    //
    if (hOut==INVALID_HANDLE_VALUE)
    {
        SetLastError(err);
    }
    return hOut;
}

