/*============================================================================
* main
============================================================================*/

#include <windows.h>
#include <stdio.h>
#include <time.h>
#include <process.h>
#include <conio.h>

#include "cyioctl.h"

// {AE18AA60-7F6A-11d4-97DD-00010229B959}
static GUID CYUSBDRV_GUID = {0xae18aa60, 0x7f6a, 0x11d4, 0x97, 0xdd, 0x0, 0x1, 0x2,
0x29, 0xb9, 0x59};


static int gbl_no_read=0;
static int gbl_no_write=0;
static int gbl_ignore_EP0_errors=0;

static volatile long gbl_readCount=0;
static volatile long gbl_writeCount=0;
static volatile long gbl_PHY_err_count=0;
static volatile long gbl_LNK_err_count=0;
static volatile long gbl_init_PHY_err_count=0;
static volatile long gbl_init_LNK_err_count=0;
static volatile long gbl_EP0_OK_count=0;

static HANDLE hDevice=INVALID_HANDLE_VALUE;
static HANDLE hStopEvent=NULL;
static volatile long gbl_stop=0;
static int iExitCode=0;

static unsigned int data1=0x00000000;
static unsigned int data2=0xFFFFFFFF;

static const size_t transferSize=64*1024; // Perform 64KB reads/writes



static void requestStop(void)
{
	InterlockedExchange(&gbl_stop,1);
    SetEvent(hStopEvent);
}
static int isStop(void)
{
	//return InterlockedCompareExchange(&gbl_stop,0,0);
	return 0!=gbl_stop;
}



void help( void)
{
    printf(
		"USAGE:  FX3USBtest [-device=<device_index>]\n"
		"        [-data=<data1>,<data2>]\n"
		"        [-no_read] [-no_write]\n"
		"        [-ignore_EP0_errors]\n"
		"where:\n"
		"  <device_index>   Device index 0... (default=0).\n"
		"  <data1>,<data2>  Repeating pattern of 32-bit data words <data1> and <data2>\n"
		"                   will be sent to FX3. Default values are 0x00000000\n"
		"                   and 0xFFFFFFFF.\n"
		"  -no_read         Reading from device is not performed.\n"
		"  -no_write        Writing to device is not performed.\n"
		"  -ignore_EP0_errors    Etron drivers fail EP 0 Control Request\n"
		"                   randomly at heavy USB3.0 read/write loads. By default,\n"
		"                   Control Request failure is considered as USB\n"
		"                   communication failure.\n"
		"                   Specifing -ignore_EP0_errors option lets to\n"
		"                   experiment with USB3.0 reads/writes also on Etron hosts.\n"
		"   NB! Prefix hexadecimal numbers with 0x\n"
		);
    exit(3);
}



extern HANDLE
OpenViaInterface(
                 LPGUID	pInterfaceGuid,
                 ULONG InstanceIndex,
                 DWORD dwDesiredAccess,                      // access mode
                 DWORD dwShareMode,                          // share mode
                 LPSECURITY_ATTRIBUTES lpSecurityAttributes, // SD
                 DWORD dwCreationDisposition,                // how to create
                 DWORD dwFlagsAndAttributes,                 // file attributes
                 HANDLE hTemplateFile                        // handle to template file
                 );



static char *GetErrorTxt( DWORD errorId )
{
    LPVOID lpMsgBuf=NULL;
    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL, errorId,
        0 /*MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT)*/,
        (LPTSTR) &lpMsgBuf,
        0, NULL);
    return (char *)lpMsgBuf;
}

static void FreeErrorTxt( char * txt)
{
    if (txt) LocalFree(txt);
}



#define BM_REQUEST_TYPE(recipent, type, direction) ((unsigned char)((((recipent) & 0x1f)<<0) |  (((type) & 0x03)<<5) | (((direction) & 0x01)<<7)))

static int FX3ReadErrors( HANDLE h, int clear_flag,
                         unsigned long long *phy_err_count,
                         unsigned long long *link_err_count,
                         unsigned int *disconnect_count,
                         unsigned int *usb_speed)
{
    unsigned int readLen=21;  // Read errors request returns 21 bytes
    DWORD nBytes=0;
    OVERLAPPED ov={0};
    PSINGLE_TRANSFER pTransfer=NULL;
    size_t size=sizeof(SINGLE_TRANSFER) + readLen;
    unsigned char *response;

    ov.hEvent=CreateEvent(NULL,TRUE,FALSE,NULL);
    if (!ov.hEvent)
    {
        fprintf(stdout, "\nERROR: Can't create event\n");
        goto err;
    }


    pTransfer=(PSINGLE_TRANSFER)malloc(size); // The size of the two-part structure
    if (!pTransfer)
    {
        fprintf(stdout, "\nERROR: Insufficient memory\n");
        goto err;
    }

    memset(pTransfer, 0, sizeof(SINGLE_TRANSFER));

    pTransfer->SetupPacket.bmRequest = BM_REQUEST_TYPE(0 /*Device */, 2 /*Vendor*/, 1 /*IN*/);
    pTransfer->SetupPacket.bRequest = 0x01;  //Read errors request
    pTransfer->SetupPacket.wValue = (USHORT)(clear_flag?1:0);
    pTransfer->SetupPacket.wIndex = (USHORT)(0);
    pTransfer->SetupPacket.wLength = readLen;
    pTransfer->SetupPacket.ulTimeOut = 30000 / 1000;
    pTransfer->reserved = 0;
    pTransfer->ucEndpointAddress = 0x00; // Control pipe
    pTransfer->IsoPacketLength = 0;
    pTransfer->BufferOffset = sizeof (SINGLE_TRANSFER);
    pTransfer->BufferLength = readLen;

    if (DeviceIoControl( h, IOCTL_ADAPT_SEND_EP0_CONTROL_TRANSFER,
        pTransfer, size, pTransfer, size, &nBytes, &ov))
    {
        // OK
    }
    else
    {
        DWORD LastError=GetLastError();
        char *msg;
        if (LastError==ERROR_IO_PENDING)
        {
			int stop;
			HANDLE hh[2]={ov.hEvent,hStopEvent};
			WaitForMultipleObjects(2,hh,FALSE,INFINITE);
			stop=isStop();
            if (stop)
            {
                WaitForMultipleObjects(1,&ov.hEvent,FALSE,1000);  // let's give about 1 second for the transfer to finish
				if (!GetOverlappedResult(h,&ov, &nBytes, FALSE))
				{
					// cancel
					BOOL (WINAPI *pCancelIoEx)(__in HANDLE hFile,__in_opt  LPOVERLAPPED lpOverlapped)=
						(BOOL (WINAPI *)(__in HANDLE,__in_opt  LPOVERLAPPED))
						GetProcAddress(GetModuleHandle("Kernel32.dll"),"CancelIoEx");

					fprintf(stdout,"\nCancelling Control pipe request\n");
					if (pCancelIoEx) (*pCancelIoEx)( h, &ov); //cancel only current write request
					else CancelIo( h);  //cancel all I/Os started by this thread
				}
            }


            if (!GetOverlappedResult(h,&ov, &nBytes, TRUE))
            {
                LastError=GetLastError();
				if (!gbl_ignore_EP0_errors || stop)
				{
					msg=GetErrorTxt(LastError);
					fprintf(stdout, "\nControl pipe DeviceIoControl failed (GetOverlappedResult error code=%lu)\n%s\n",
						(long unsigned)LastError, msg?msg:"" );
					FreeErrorTxt(msg);
				}
                goto err;
            }
        }
        else
        {
			if (!gbl_ignore_EP0_errors)
			{
				msg=GetErrorTxt(LastError);
				fprintf(stdout, "\nControl pipe DeviceIoControl failed (error code=%lu)\n%s\n",
					(long unsigned)LastError, msg?msg:"" );
				FreeErrorTxt(msg);
			}
            goto err;
        }
    }

    nBytes=nBytes>sizeof(SINGLE_TRANSFER)?nBytes-sizeof(SINGLE_TRANSFER):0;
    if (nBytes<readLen)
    {
        fprintf(stdout, "\nDevice returned less info than expected (%lu bytes of %lu)\n", (unsigned long)nBytes,(unsigned long)readLen );
        goto err;
    }

    // process response
    response=(unsigned char *)pTransfer+sizeof(SINGLE_TRANSFER);

    // 64-bit PHY errors count in LE
    if (phy_err_count)
    {
        *phy_err_count=((unsigned long long)response[0]<<0)+((unsigned long long)response[1]<<8)
            +((unsigned long long)response[2]<<16)+((unsigned long long)response[3]<<24)
            +((unsigned long long)response[4]<<32)+((unsigned long long)response[5]<<40)
            +((unsigned long long)response[6]<<48)+((unsigned long long)response[7]<<56);
    }
    response+=8;

    // 64-bit LINK errors count in LE
    if (link_err_count)
    {
        *link_err_count=((unsigned long long)response[0]<<0)+((unsigned long long)response[1]<<8)
            +((unsigned long long)response[2]<<16)+((unsigned long long)response[3]<<24)
            +((unsigned long long)response[4]<<32)+((unsigned long long)response[5]<<40)
            +((unsigned long long)response[6]<<48)+((unsigned long long)response[7]<<56);
    }
    response+=8;

    // 32-bit USB Disconnect/Reset count in LE
    if (disconnect_count)
    {
        *disconnect_count=((unsigned int)response[0]<<0)+((unsigned int)response[1]<<8)
            +((unsigned int)response[2]<<16)+((unsigned int)response[3]<<24);
    }
    response+=4;

    // 8-bit USB speed
    if (usb_speed)
    {
        *usb_speed=((unsigned int)response[0]<<0);
    }
    response+=1;

    if (pTransfer) free(pTransfer);
    if (ov.hEvent) CloseHandle(ov.hEvent);

    return 0;
err:
    if (pTransfer) free(pTransfer);
    if (ov.hEvent) CloseHandle(ov.hEvent);
    return 1;
}


static unsigned __stdcall UI_thread( void *dummy )
{
    unsigned long c_read=0, c_write=0, c_prev_read=0, c_prev_write=0;
	unsigned long c;
	unsigned long long c_PHY=0, c_LNK=0, c_EP0_OK=0;
    double speed_read=0, speed_write=0;
    DWORD t, t_prev, t_print, t_start;
	unsigned int EP0_activity=10;

	unsigned long long tl_current, tl_start;

    fprintf(stdout,"\nPress ESC for exit\n");

    t=GetTickCount();
	tl_start=t;
	tl_current=t;
	t_prev=t;
    t_print=t;

    while(!isStop())
    {
        if (_kbhit())
        {
            switch (/*getchar()*/ _getch())
            {
            case 27:
                goto ret;
            default:
                break;
            }
        }

        t=GetTickCount();
		if (t<(tl_current & 0xFFFFFFFFULL))
		{
			tl_current += (1ULL<<32);
		}
		tl_current =  (tl_current & ~ 0xFFFFFFFFULL) | t;

		// refresh display twice per second
        if (t-t_print>=500)
        {
			unsigned char time_buf[sizeof(unsigned long)*3+1+8+1];
            c_read=InterlockedExchangeAdd(&gbl_readCount,(unsigned long) 0);
            c_write=InterlockedExchangeAdd(&gbl_writeCount,(unsigned long) 0);
            if (t-t_prev>1000)
            {
                speed_read=1e3*(c_read-c_prev_read)/(t-t_prev);
                speed_write=1e3*(c_write-c_prev_write)/(t-t_prev);
                t_prev=t;
                c_prev_read=c_read;
				c_prev_write=c_write;
            }
			c=InterlockedExchangeAdd(&gbl_PHY_err_count,(unsigned long) 0);

			if (c < (unsigned long)c_PHY)
			{
				c_PHY+=0x100000000ULL;
			}
			c_PHY=(c_PHY & ~0xFFFFFFFFULL) | c;
			c_PHY-=gbl_init_PHY_err_count;

			c=InterlockedExchangeAdd(&gbl_LNK_err_count,(unsigned long) 0);
			if (c < (unsigned long)c_LNK)
			{
				c_LNK+=0x100000000ULL;
			}
			c_LNK=(c_LNK & ~0xFFFFFFFFULL) | c;
			c_LNK-=gbl_init_LNK_err_count;

			c=InterlockedExchangeAdd(&gbl_EP0_OK_count,(unsigned long) 0);
			if (c_EP0_OK!=c)
			{
				EP0_activity=10;
				c_EP0_OK=c;
			}
			else if (EP0_activity)
			{
				EP0_activity-=1;
			}


			{
                unsigned long days, hours, minutes, seconds;
                unsigned long long duration=tl_current-tl_start;

                days=(unsigned long)(duration/(1000ULL*60*60*24));
                duration -= days*(1000ULL*60*60*24);
                hours=(unsigned long)(duration/(1000ULL*60*60));
                duration -= hours*(1000ULL*60*60);
                minutes=(unsigned long)(duration/(1000ULL*60));
                duration -= minutes*(1000ULL*60);
                seconds=(unsigned long)(duration/(1000ULL));
                //duration -= seconds*(1000ULL);

                if (days)
                {
                    if (0>_snprintf(time_buf,sizeof(time_buf),"%lud%02lu:%02lu:%02lu", (unsigned long)days,(unsigned long)hours, (unsigned long)minutes, (unsigned long)seconds))
					{
						// Program internal error, time_buf too small
						time_buf[sizeof(time_buf)-1]=0;
					}

                }
                else
                {
                    if (0>_snprintf(time_buf,sizeof(time_buf),"%02lu:%02lu:%02lu", (unsigned long)hours, (unsigned long)minutes, (unsigned long)seconds))
					{
						// Program internal error, time_buf too small
						time_buf[sizeof(time_buf)-1]=0;
					}
                }
				
			}


            fprintf(stdout,"\r%s  Read/Write=%5.1lf/%5.1lfMB/s  Errors %s=%llu/%llu\r", 
                    time_buf,
                    speed_read*(1./(1024*1024)), speed_write*(1./(1024*1024)),
					EP0_activity?"PHY/LNK":"phy/lnk", // let's indicate with letter charcaters that something is wrong with updating error counters, i.e. with device EP0 requests
                    (unsigned long long)c_PHY, (unsigned long long)c_LNK
					);
            t_print=t;
        }
        // Sleep
        WaitForSingleObject(hStopEvent,100);
    }
ret:
    fprintf(stdout,"\n");
	requestStop();
    return 0;
}

static unsigned __stdcall Err_thread( void *dummy )
{
	// Query error counters
	while(!isStop())
	{
		unsigned __int64 link_err_count,phy_err_count;
		int err;

		err=FX3ReadErrors( hDevice, 0, &phy_err_count, &link_err_count,	NULL, NULL);
		if (!err)
		{
			//NB! 64-bit error counters returned from FX3 FX3GPIFnoise application
			//    may be randomly incorrect due to missing synchronization.
			//    Let's use only lower 32-bits that are guaranteed to be correct.
			InterlockedExchange(&gbl_PHY_err_count,(unsigned long) phy_err_count);
			InterlockedExchange(&gbl_LNK_err_count,(unsigned long) link_err_count);

			// Control Request succeeded, increment relevant counter
			InterlockedIncrement(&gbl_EP0_OK_count);

		}
		else
		{
			if (!gbl_ignore_EP0_errors)
			{
				if (!isStop())
				{
					// Control Request seems to cause failure
					fprintf(stdout,
						"Note, if your host uses Etron USB3.0 driver, consider running\n"
						"FX3USBtest  with -ignore_EP0_errors option\n"
						"Type   FX3USBtest -help    for getting help on options\n");
				}
				iExitCode=3;
				goto ret;
			}
		}
		
		// Sleep a bit
		WaitForSingleObject(hStopEvent,1000);
	}
ret:
	requestStop();
	return (0);
}


static unsigned __stdcall Rd_thread( void *dummy )
{
	unsigned char *dataBuf=NULL;
	SINGLE_TRANSFER *pTransfer=NULL;
	OVERLAPPED ov={0};
	HANDLE hh[2];

	ov.hEvent=CreateEvent(NULL,TRUE,FALSE,NULL);
	if (!ov.hEvent)
	{
        fprintf(stdout,"\nERROR: failed to create Event\n");
        iExitCode=3;
        goto ret;
	}

    dataBuf=(unsigned char *)malloc(transferSize);
    if (!dataBuf)
    {
        fprintf(stdout,"\nERROR: Insufficient memory\n");
        iExitCode=3;
        goto ret;
    }

    pTransfer=(SINGLE_TRANSFER *)malloc(sizeof (SINGLE_TRANSFER));
    if (!pTransfer)
    {
        fprintf( stdout,"\nERROR: Insufficient memory\n" );
        iExitCode=3;
        goto ret;
    }
    memset(pTransfer,0,sizeof (SINGLE_TRANSFER));
    pTransfer->ucEndpointAddress = /*Address*/0x81;  //Bulk IN port
    pTransfer->IsoPacketLength = 0;
    pTransfer->BufferOffset = 0;
    pTransfer->BufferLength = 0;


    hh[0]=ov.hEvent;
    hh[1]=hStopEvent;
    while(!isStop())
    {
        DWORD dwReturnBytes;

        if (DeviceIoControl (hDevice, IOCTL_ADAPT_SEND_NON_EP0_DIRECT, pTransfer, sizeof (SINGLE_TRANSFER),
            dataBuf, transferSize, &dwReturnBytes, &ov))
        {
            // OK
        }
        else
        {
            DWORD LastError = GetLastError();
            if (ERROR_IO_PENDING==LastError)
            {
				long stop;
                // wait until DeviceIoControl finishes or user stops
                WaitForMultipleObjects(2,hh,FALSE,INFINITE);
				stop=isStop();
                if (stop)
                {
					WaitForMultipleObjects(1,&ov.hEvent,FALSE,1000);  // let's give about 1 second for the transfer to finish
					if (!GetOverlappedResult(hDevice,&ov, &dwReturnBytes, FALSE))
					{
						// cancel
						BOOL (WINAPI *pCancelIoEx)(__in HANDLE hFile,__in_opt  LPOVERLAPPED lpOverlapped)=
							(BOOL (WINAPI *)(__in HANDLE,__in_opt  LPOVERLAPPED))
							GetProcAddress(GetModuleHandle("Kernel32.dll"),"CancelIoEx");
						fprintf(stdout,"Cancelling read operation\n");

						if (pCancelIoEx) (*pCancelIoEx)( hDevice, &ov ); //cancel only current write request
						else CancelIo( hDevice );  //cancel all I/Os started by this thread
					}
				}

                // wait until DeviceIoControl finishes
                if (!GetOverlappedResult(hDevice, &ov, &dwReturnBytes, TRUE))
                {
                    LastError = GetLastError();
                    if (/*!stop*/1)  // display all errors, incl. ERROR_CANCELLED as well if operation was canceled due to exit
                    {
                        char *msg=GetErrorTxt(LastError);
						fprintf( stdout, "\nRead from device failed (GetOverlappedResult error code=%lu)\n%s\n",
                            (long unsigned)LastError, msg?msg:"" );
                        FreeErrorTxt(msg);
                    }
                    if (!stop)
                    {
                        iExitCode=2;
                    }
                    goto ret;
                }
            }
            else
            {
                char *msg=GetErrorTxt(LastError);
                fprintf( stdout, "\nRead from device failed (error code=%lu)\n%s\n",
                    (long unsigned)LastError, msg?msg:"" );
                FreeErrorTxt(msg);
                iExitCode=2;
                goto ret;
            }

        }

        // update bytes counter
        InterlockedExchangeAdd(&gbl_readCount,(unsigned long) dwReturnBytes);
    }

ret:
	if (dataBuf) free(dataBuf);
	if (pTransfer) free(pTransfer);
	if (ov.hEvent) CloseHandle(ov.hEvent);

	if (!isStop())
	{
		requestStop();
	}
	return (0);
}


static unsigned __stdcall Wr_thread( void *dummy )
{
	size_t dataBufSize;
	unsigned char *dataBuf=NULL;
	SINGLE_TRANSFER *pTransfer=NULL;
	OVERLAPPED ov={0};
	HANDLE hh[2];

	ov.hEvent=CreateEvent(NULL,TRUE,FALSE,NULL);
	if (!ov.hEvent)
	{
        fprintf(stdout,"\nERROR: failed to create Event\n");
        iExitCode=3;
        goto ret;
	}


    dataBufSize=(transferSize+(sizeof(unsigned int)-1))/sizeof(unsigned int)*sizeof(unsigned int);
    dataBuf=malloc(dataBufSize);
    if (!dataBuf)
    {
        fprintf(stdout,"\nERROR: Insufficient memory\n");
        iExitCode=3;
        goto ret;
    }

	// fill data buffer with data pattern (by default toggling) values
	{
		size_t i;
		for (i=0; i<dataBufSize/sizeof(unsigned int); i++)
		{
			if (0==(i & 1)) ((unsigned int*)dataBuf)[i]=data1;
			else ((unsigned int*)dataBuf)[i]=data2;
		}
	}

    pTransfer=(SINGLE_TRANSFER *)malloc(sizeof (SINGLE_TRANSFER));
    if (!pTransfer)
    {
        fprintf(stdout,"\nERROR: Insufficient memory\n");
        iExitCode=3;
        goto ret;
    }
    memset(pTransfer,0,sizeof (SINGLE_TRANSFER));
    pTransfer->ucEndpointAddress = /*Address*/0x01;
    pTransfer->IsoPacketLength = 0;
    pTransfer->BufferOffset = 0;
    pTransfer->BufferLength = 0;

    hh[0]=ov.hEvent;
    hh[1]=hStopEvent;

    while(!isStop())
    {
        DWORD dwReturnBytes;

        if (DeviceIoControl(hDevice, IOCTL_ADAPT_SEND_NON_EP0_DIRECT, pTransfer, sizeof (SINGLE_TRANSFER),
            dataBuf, transferSize, &dwReturnBytes, &ov))
        {
            // OK
        }
        else
        {
            DWORD LastError = GetLastError();
            if (ERROR_IO_PENDING==LastError)
            {
				int stop;
                // wait until DeviceIoControl finishes or user stops
                WaitForMultipleObjects(2,hh,FALSE,INFINITE);
				stop=isStop();
                if (stop)
                {
					WaitForMultipleObjects(1,&ov.hEvent,FALSE,1000);  // let's give about 1 second for the transfer to finish
					if (!GetOverlappedResult(hDevice,&ov, &dwReturnBytes, FALSE))
					{
						// cancel
						BOOL (WINAPI *pCancelIoEx)(__in HANDLE hFile,__in_opt  LPOVERLAPPED lpOverlapped)=
							(BOOL (WINAPI *)(__in HANDLE,__in_opt  LPOVERLAPPED))
							GetProcAddress(GetModuleHandle("Kernel32.dll"),"CancelIoEx");
						
						fprintf(stdout,"Cancelling write operation\n");
						if (pCancelIoEx) (*pCancelIoEx)( hDevice, &ov ); //cancel only current write request
						else CancelIo( hDevice );  //cancel all I/Os started by this thread
					}
                }

                // wait until DeviceIoControl finishes
                if (!GetOverlappedResult(hDevice, &ov, &dwReturnBytes, TRUE))
                {
                    LastError=GetLastError();
                    if (/*!stop*/1)  // display all errors, incl. ERROR_CANCELLED if operation was canceled due to exit
                    {
                        char *msg=GetErrorTxt(LastError);
                        fprintf(stdout, "\nWrite to device failed (GetOverlappedResult error code=%lu)\n%s\n",
                            (long unsigned)LastError, msg?msg:"" );
                        FreeErrorTxt(msg);
                    }
                    if (!stop)
                    {
                        iExitCode=2;
                    }
                    goto ret;
                }
            }
            else
            {
                char *msg=GetErrorTxt(LastError);
                fprintf(stdout, "\nWrite to device failed (error code=%lu)\n%s\n",
                    (long unsigned)LastError, msg?msg:"" );
                FreeErrorTxt(msg);
                iExitCode=2;
                goto ret;
            }

        }

        InterlockedExchangeAdd(&gbl_writeCount,(unsigned long) dwReturnBytes);
    }

ret:
	
    if (dataBuf) free(dataBuf);
    if (pTransfer) free(pTransfer);
    if (ov.hEvent) CloseHandle(ov.hEvent);
	if (!isStop())
	{
		requestStop();;
	}

    return( 0 );
}


int main( int iArgc, char** ppzArgv )
{
    LPGUID	pGuid=&CYUSBDRV_GUID;
    int     i;
    int iDevice=0;
	int ignore_EP0_errors=0;
	unsigned int threadID;
	uintptr_t hUI_thread=0;
    uintptr_t hRd_thread=0;
	uintptr_t hWr_thread=0;
	uintptr_t hErr_thread=0;



    printf( "FX3USBtest version. 1.3\n" );

    for ( i = 1; i < iArgc ; ++i )
    {
        if (0)
        {
        }
        else if ( !_strnicmp( ppzArgv[ i ], "-device", 7 ) )
        {
            unsigned long n;
            unsigned char c;
            if (0
                || 1==sscanf(ppzArgv[ i ]+7," = 0x%lX %c", &n, &c)
                || 1==sscanf(ppzArgv[ i ]+7," = 0X%lX %c", &n, &c)
                || 1==sscanf(ppzArgv[ i ]+7," = %lu %c", &n, &c))
            {
                iDevice    = (unsigned int)n;
                printf( "Device index = %u\n", (unsigned int)iDevice );
            }
            else help();
        }
        else if ( !_strnicmp( ppzArgv[ i ], "-data", 5 ) )
        {
            unsigned long n1,n2;
            unsigned char c;
            if (0
                || 2==sscanf(ppzArgv[ i ]+5," = 0x%lX , 0x%lX %c", &n1,&n2, &c)
                || 2==sscanf(ppzArgv[ i ]+5," = 0x%lX , 0X%lX %c", &n1,&n2, &c)
                || 2==sscanf(ppzArgv[ i ]+5," = 0x%lX , %lu %c", &n1,&n2, &c)
				|| 2==sscanf(ppzArgv[ i ]+5," = 0X%lX , 0x%lX %c", &n1,&n2, &c)
				|| 2==sscanf(ppzArgv[ i ]+5," = 0X%lX , 0X%lX %c", &n1,&n2, &c)
				|| 2==sscanf(ppzArgv[ i ]+5," = 0X%lX , %lu %c", &n1,&n2, &c)
				|| 2==sscanf(ppzArgv[ i ]+5," = %lu , 0x%lX %c", &n1,&n2, &c)
				|| 2==sscanf(ppzArgv[ i ]+5," = %lu , 0X%lX %c", &n1,&n2, &c)
				|| 2==sscanf(ppzArgv[ i ]+5," = %lu , %lu %c", &n1,&n2, &c)
				)
            {
                data1 = (unsigned long)n1;
				data2 = (unsigned long)n2;
                printf( "Data pattern = 0x%08lX,0x%08lX\n", (unsigned long)data1,(unsigned long)data2 );
            }
            else help();
        }
		else if ( !_strnicmp( ppzArgv[ i ], "-no_read", 8 ) || !_strnicmp( ppzArgv[ i ], "-no-read", 8 ) )
        {
            gbl_no_read=1;
        }
		else if ( !_strnicmp( ppzArgv[ i ], "-no_write", 9 ) || !_strnicmp( ppzArgv[ i ], "-no-write", 9 ) )
        {
            gbl_no_write=1;
        }
		else if ( !_strnicmp( ppzArgv[ i ], "-ignore_EP0_errors", 18 ) || !_strnicmp( ppzArgv[ i ], "-ignore_EP0_errors", 18 ) )
        {
            ignore_EP0_errors=1;
        }
        else
        {
            help();
        }
    }

	hStopEvent=CreateEvent(NULL,TRUE,FALSE,NULL);
	if (!hStopEvent)
	{
        fprintf( stdout, "\nERROR:  Cannot create event\n" );
        iExitCode=3;
        goto ret;
	}

    hDevice=OpenViaInterface(
        pGuid,
        iDevice,
        GENERIC_READ | GENERIC_WRITE,
        FILE_SHARE_READ | FILE_SHARE_WRITE,
        NULL, // no SECURITY_ATTRIBUTES structure
        OPEN_EXISTING,
        FILE_FLAG_OVERLAPPED,
        NULL);
    if (INVALID_HANDLE_VALUE==hDevice)
    {
        fprintf( stdout, "\nERROR:  Cannot open device %d\n", iDevice );
        iExitCode=3;
        goto ret;
    }


	hUI_thread=_beginthreadex(
        NULL,		//security
        0,			//stack size
        UI_thread,	//start address
        0,		//argument list
        0,			//initial state of new thread,
        &threadID	//thread identifier
        );
    if (0==hUI_thread)
    {
        fprintf( stdout, "\nERROR: _beginthreadex failed, error code=%u\n",(unsigned int)errno );
        iExitCode=3;
        goto ret;
    }

	// Clear device error counters
#if 0
	if (FX3ReadErrors( hDevice, 1, NULL, NULL,	NULL, NULL))
	{
		iExitCode=2;
		goto ret;
	}
#else
	{
		unsigned __int64 phy_err_count, link_err_count;
		unsigned int disconnect_count;
		int speed=0;
		if (FX3ReadErrors( hDevice, 0, &phy_err_count, &link_err_count,	&disconnect_count, &speed))
		{
			iExitCode=2;
			goto ret;
		}
		gbl_init_PHY_err_count=(unsigned long) phy_err_count;
		gbl_init_LNK_err_count=(unsigned long) link_err_count;
		fprintf(stdout,"Device has been disconnected %lu times\nDevice is in %s mode\nDevice internal error counters at program start: PHY/LNK=%llu/%llu\nProgram treats these values as initial zero point to continue counting on\n", 
			(unsigned long)disconnect_count,
			speed==3?"USB3 Super Speed":(speed==2?"USB2 High speed":(speed==1?"USB2 Full speed":"unknown speed")),
			(unsigned long long)gbl_init_PHY_err_count,
			(unsigned long long)gbl_init_LNK_err_count);
	}
#endif
    gbl_ignore_EP0_errors=ignore_EP0_errors;


	// Start threads
	hErr_thread=_beginthreadex(
		NULL,		//security
		0,			//stack size
		Err_thread,	//start address
		0,		//argument list
		0,			//initial state of new thread,
		&threadID	//thread identifier
	);
	if (0==hErr_thread)
	{
		fprintf( stdout, "\nERROR: _beginthreadex failed, error code=%u\n",(unsigned int)errno );
		iExitCode=3;
		goto ret;
	}



	if (!gbl_no_write)
	{
		hWr_thread=_beginthreadex(
			NULL,		//security
			0,			//stack size
			Wr_thread,	//start address
			0,		//argument list
			0,			//initial state of new thread,
			&threadID	//thread identifier
		);
		if (0==hWr_thread)
		{
			fprintf( stdout, "\nERROR: _beginthreadex failed, error code=%u\n",(unsigned int)errno );
			iExitCode=3;
			goto ret;
		}
	}

	if (!gbl_no_read)
	{
		hRd_thread=_beginthreadex(
			NULL,		//security
			0,			//stack size
			Rd_thread,	//start address
			0,		//argument list
			0,			//initial state of new thread,
			&threadID	//thread identifier
		);
		if (0==hRd_thread)
		{
			fprintf( stdout, "\nERROR: _beginthreadex failed, error code=%u\n",(unsigned int)errno );
			iExitCode=3;
			goto ret;
		}
	}

	// wait for stop
	while(!isStop())
	{
		WaitForSingleObject(hStopEvent,INFINITE);
	}


ret:
	requestStop();

	if (hUI_thread)
	{
		WaitForSingleObject((HANDLE)hUI_thread,INFINITE);
		CloseHandle((HANDLE)hUI_thread);
	}
	if (hRd_thread)
	{
		WaitForSingleObject((HANDLE)hRd_thread,INFINITE);
		CloseHandle((HANDLE)hRd_thread);
	}
	if (hWr_thread)
	{
		WaitForSingleObject((HANDLE)hWr_thread,INFINITE);
		CloseHandle((HANDLE)hWr_thread);
	}
	if (hErr_thread)
	{
		WaitForSingleObject((HANDLE)hErr_thread,INFINITE);
		CloseHandle((HANDLE)hErr_thread);
	}
    if (hDevice!=INVALID_HANDLE_VALUE) CloseHandle(hDevice);

    return( iExitCode );
}


